#include <type_traits>
#include <cassert>

namespace op {

// minimal complete implementation: operator ==
struct Eq { };

// minimal complete implementation: operator<
struct Ord : Eq { };

// a SFINAE helper used in return type to make sure the operator can be
// instantiated only for type that inherits from a given tag (Eq, Ord, …)
template< typename T, typename Trait, typename R >
using IsTrait = std::enable_if_t< std::is_base_of< Trait,
                std::remove_reference_t< T > >::value, R >;

template< typename T, typename R = bool >
using IsEq = IsTrait< T, Eq, R >;

template< typename T, typename R = bool >
using IsOrd = IsTrait< T, Ord, R >;

// everything that inherits (indirectly) from Eq can use this operator !=
template< typename T >
IsEq< T > operator!=( const T &a, const T &b ) { return !(a == b); }

// everything that inherits from Ord can use these comparison operators,
// the goal is to implement them using one comparison
template< typename T >
IsOrd< T > operator>( const T &a, const T &b ) { return b < a; }

template< typename T >
IsOrd< T > operator>=( const T &a, const T &b ) { return !(a < b); }

template< typename T >
IsOrd< T > operator<=( const T &a, const T &b ) { return !(b < a); }

// Since C++17 non-type template arguments can also be arguments of previously
// unknown types and the compiler can derive these types. This can be used to
// implement compile-time list of values, such as member value pointers which
// can then be used to implement pairwise comparison.
template< auto... vals >
struct CompileTimeList { };

// Tags for pairwise comparison and lexicographical ordering;
// apart from inheriting from them, the class must define the member type
// MemberPtrTuple which is a typedef of CompileTimeList of member pointers
// that should be used for the comparison (in the order used for the
// lexicographical ordering). No operators are required.
struct PairwiseEq : Eq { };
struct LexOrd : Ord, PairwiseEq { };

template< typename T, auto ... ptr >
bool tupleEq( const T &a, const T &b, CompileTimeList< ptr... > /*unused*/ ) {
    // See operator== first.
    // Pattern matching on CompileTimeList is used to extract member pointers
    // from it and these pointers are them used to implement pairwise
    // comparison using fold expression.
    return ( (a.*ptr == b.*ptr) && ... );
}

template< typename T >
IsTrait< T, PairwiseEq, bool > operator==( const T &a, const T &b ) {
    // pass the MemberPtrTuple by value to the helper function so it can
    // pattern match on it
    return tupleEq( a, b, typename T::MemberPtrTuple() );
}

template< typename T, auto ptr, auto ... ptrs >
bool lexLess( const T &a, const T &b, CompileTimeList< ptr, ptrs... > /*unused*/ ) {
    // Similar to tupleEq, but we have to use recursion and specialization for
    // empty CompileTimeList as a base case (the expression is too complex to
    // be written as fold).
    return (a.*ptr) < (b.*ptr)
        || (a.*ptr == b.*ptr && lexLess( a, b, CompileTimeList< ptrs... >() ));
}

template< typename T >
bool lexLess( const T &/*unused*/, const T &/*unused*/, CompileTimeList<> /*unused*/ ) {
    // base case, all things were equal, so the two classes are equal, not a < b
    return false;
}

template< typename T >
IsTrait< T, LexOrd, bool > operator<( const T &a, const T &b ) {
    // again use the helper to extract member pointers
    return lexLess( a, b, typename T::MemberPtrTuple() );
}

} // namespace op

class Test : op::LexOrd {

    int a, b, c;
    long d;
  public:

    Test( int a, int b, int c, long d ) : a( a ), b( b ), c( c ), d( d ) { }

    // lists all members which should be used for comparison/ordering
    using MemberPtrTuple = op::CompileTimeList< &Test::a, &Test::b, &Test::c, &Test::d >;
};

int main() {
    Test t1 = { 1, 1, 1, 1 },
         t2 = { 1, 1, 2, 2 };

    assert( t1 == t1 );
    assert( t1 != t2 );
    assert( t1 <= t1 );
    assert( !(t1 < t1) );
    assert( t1 < t2 );
}
